/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.novabpm.identity.auth;

import java.security.Principal;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.management.remote.JMXPrincipal;
import javax.security.auth.Destroyable;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import org.ow2.novabpm.identity.IdentityServiceOp;
import org.ow2.novabpm.util.Misc;

/**
 * @author "Pierre Vigneras"
 * @date Dec 3, 2007
 */
public class IdLoginModule implements LoginModule {
  private static final String NAME_PROMPT = "Name: ";

  private static final String PASSWORD_PROMPT = "Password: ";

  private static final String JAVAX_SECURITY_AUTH_LOGIN_PASSWORD = "javax.security.auth.login.password";

  private static final String JAVAX_SECURITY_AUTH_LOGIN_NAME = "javax.security.auth.login.name";

  /**
   * Property key used to define the authentication service to use. Defined to
   * be "service.impl". Property Value should be the full name of a class that
   * implements the {@link IdentityServiceOp} interface.
   */
  public static final String AUTH_SERVICE_URL = "service.impl";

  /**
   * Property key used to define service instanciation parameters. Defined to be
   * "service.parameters". Property Value can be any String.
   */
  public static final String SERVICE_PARAMS_OPTION_NAME = "service.parameters";

  /**
   * Property key for the debug flag. Defined to be "debug".
   *
   * Property Value. If set, should be either "true" or "false". Default is
   * "false".
   */
  public static final String DEBUG_OPTION_NAME = "debug";

  private static final String JMX_URL_OPTION_NAME = "jmx.url";

  private static final String AUTH_SERVICE_NAME_OPTION_NAME = "auth.service.name";

  private AuthenticationServiceMBean service;

  private Subject subject = null;

  private CallbackHandler callbackHandler = null;

  private Map<String, Object> sharedState;

  private Map<String, Object> options = null;

  private boolean debug = false;

  /**
   * Roles of the principal
   */
  private ArrayList roles = null;

  private String id;

  /**
   * Initialize this LoginModule. This method is called by the LoginContext
   * after this LoginModule has been instantiated. The purpose of this method is
   * to initialize this LoginModule with the relevant information. If this
   * LoginModule does not understand any of the data stored in sharedState or
   * options parameters, they can be ignored.
   *
   * @param subject
   *            the Subject to be authenticated.
   * @param callbackHandler
   *            a CallbackHandler for communicating with the end user (prompting
   *            for usernames and passwords, for example).
   * @param sharedState
   *            state shared with other configured LoginModules.
   * @param options
   *            options specified in the login Configuration for this particular
   *            LoginModule.
   */
  public void initialize(Subject subject, CallbackHandler callbackHandler,
      Map<String, ?> sharedState, Map<String, ?> options) {

    this.subject = subject;
    this.callbackHandler = callbackHandler;
    this.sharedState = (Map<String, Object>) sharedState;
    this.options = (Map<String, Object>) options;
    final String debugFlag = (String) options.get(DEBUG_OPTION_NAME);
    if (debugFlag != null) {
      debug = Boolean.valueOf(debugFlag);
    }
  }

  /**
   * Method to authenticate a Subject (phase 1). The implementation of this
   * method authenticates a Subject. For example, it may prompt for Subject
   * information such as a username and password and then attempt to verify the
   * password. This method saves the result of the authentication attempt as
   * private state within the LoginModule.
   *
   * @return true if the authentication succeeded, or false if this LoginModule
   *         should be ignored.
   * @throws LoginException
   *             if the authentication fails
   */
  public boolean login() throws LoginException {
    if (debug)
      System.err.println("[" + IdLoginModule.class.getName()
          + "] login() - preparing - step 1");
    if (service == null) {
      if (debug)
        System.err.println("[" + IdLoginModule.class.getName()
            + "] login() - parsing options");
      final String jmxUrl = (String) options.get(JMX_URL_OPTION_NAME);
      if (jmxUrl == null)
        throw new MissingOptionException(JMX_URL_OPTION_NAME);
      final String authServiceBindingName = (String) options
          .get(AUTH_SERVICE_NAME_OPTION_NAME);
      if (authServiceBindingName == null)
        throw new MissingOptionException(AUTH_SERVICE_NAME_OPTION_NAME);
      try {
        this.service = Misc.getMBeanProxy(AuthenticationServiceMBean.class,
            jmxUrl, authServiceBindingName);
      } catch (Exception e) {
        final LoginException le = new LoginException();
        le.initCause(e);
        throw le;
      }
    }
    try {
      String name = (String) sharedState.get(JAVAX_SECURITY_AUTH_LOGIN_NAME);
      String password = (String) sharedState
          .get(JAVAX_SECURITY_AUTH_LOGIN_PASSWORD);
      final List<Callback> callbacks = new ArrayList<Callback>();
      final NameCallback nameCallback = new NameCallback(NAME_PROMPT);
      final PasswordCallback passwordCallback = new PasswordCallback(
          PASSWORD_PROMPT, false);
      if (name == null)
        callbacks.add(nameCallback);
      if (password == null)
        callbacks.add(passwordCallback);
      if (!callbacks.isEmpty()) {
        if (debug)
          System.err.println("[" + IdLoginModule.class.getName()
              + "] login() - callback - step 2");
        callbackHandler.handle(callbacks.toArray(new Callback[0]));
        if (name == null)
          name = nameCallback.getName();
        if (password == null) {
          password = new String(passwordCallback.getPassword());
          passwordCallback.clearPassword();
        }
      }
      if (debug)
        System.err.println("[" + IdLoginModule.class.getName()
            + "] login() - authenticating - step 3");
      this.id = service.authenticateUser(name, password);
      // TODO: what is Step 4 then?!
      if (debug)
        System.err.println("[" + IdLoginModule.class.getName()
            + "] login() - storing data - step 5");
      sharedState.put(JAVAX_SECURITY_AUTH_LOGIN_NAME, name);
      sharedState.put(JAVAX_SECURITY_AUTH_LOGIN_PASSWORD, password);
      if (debug)
        System.err.println("[" + IdLoginModule.class.getName()
            + "] login() - returning - step 6");
      if (id == null)
        throw new FailedLoginException("User unknown or wrong password!");
      return true;
    } catch (Exception e) {
      final LoginException le = new LoginException();
      le.initCause(e);
      throw le;
    }
  }

  /**
   * Method to commit the authentication process (phase 2). This method is
   * called if the LoginContext's overall authentication succeeded (the relevant
   * REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules succeeded). If
   * this LoginModule's own authentication attempt succeeded (checked by
   * retrieving the private state saved by the login method), then this method
   * associates relevant Principals and Credentials with the Subject located in
   * the LoginModule. If this LoginModule's own authentication attempted failed,
   * then this method removes/destroys any state that was originally saved.
   *
   * @return true if this method succeeded, or false if this LoginModule should
   *         be ignored.
   * @throws LoginException
   *             if the commit fails
   */
  public boolean commit() throws LoginException {
    final String name = (String) sharedState
        .get(JAVAX_SECURITY_AUTH_LOGIN_NAME);
    if (id == null)
      throw new FailedLoginException("User unknown or wrong password!");
    final Set<Principal> principals = subject.getPrincipals();
    // We use JMXPrincipal as it is the single Principal implementation available 
    // in the whole J2SE 5.
    principals.add(new JMXPrincipal(id));
    return true;
  }

  /**
   * Method to abort the authentication process (phase 2). This method is called
   * if the LoginContext's overall authentication failed. (the relevant
   * REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules did not succeed).
   * If this LoginModule's own authentication attempt succeeded (checked by
   * retrieving the private state saved by the login method), then this method
   * cleans up any state that was originally saved.
   *
   * @return true if this method succeeded, or false if this LoginModule should
   *         be ignored.
   * @throws LoginException
   *             if the abort fails
   */
  public boolean abort() throws LoginException {
    if (debug)
      System.err.println("[" + IdLoginModule.class.getName() + "] abort()");
    if (id == null)
      return false;
    subject = null;
    id = null;
    roles = null;
    return true;
  }

  /**
   * Method which logs out a Subject. An implementation of this method might
   * remove/destroy a Subject's Principals and Credentials.
   *
   * @return true if this method succeeded, or false if this LoginModule should
   *         be ignored.
   * @throws LoginException
   *             if the logout fails
   */
  public boolean logout() throws LoginException {
    if (id != null) {
      if (debug)
        System.err.println("[" + IdLoginModule.class.getName()
            + "] logout() - removing principals");
      // Remove only principals added by our commit method
      final Set<Principal> principals = subject.getPrincipals();
      for (Principal p : principals) {
        if (p instanceof BonitaPrincipal) {
          if (debug)
            System.err.println("[" + IdLoginModule.class.getName()
                + "] logout() - removing principal: " + p);
          principals.remove(p);
        }
      }
      if (debug)
        System.err.println("[" + IdLoginModule.class.getName()
            + "] logout() - destroying/removing credentials");
      // Remove/destroy only credentials added by our commit method
      final Set<Object> credentials = subject.getPublicCredentials();
      for (Object o : credentials) {
        if (o instanceof Destroyable) {
          if (debug)
            System.err.println("[" + IdLoginModule.class.getName()
                + "] logout() - destroying credential: " + o);
          // Bug: only from this module !!
          // ((Destroyable) o).destroy();
        }
        if (!subject.isReadOnly()) {
          if (debug)
            System.err.println("[" + IdLoginModule.class.getName()
                + "] logout() - removing credential: " + o);
          credentials.remove(o);
        }
      }
    }
    return true;
  }

  /**
   * @author "Pierre Vigneras"
   * @date Dec 3, 2007
   */
  public class MissingOptionException extends LoginException {
    /**
     * @param msg
     */
    public MissingOptionException(String msg) {
      super("Missing options: " + msg);
    }
  }
}
